//
// This version of the huffman coder applies strictly to the interface defined by the abstract entropy coder.
//

import java.io.*;
import java.util.*;

class HuffmanCoder extends EntropyCoder
{
  SymbolStatistics SymStat;
  GrowableIntArray Symbols;
  GrowableIntArray Contexts;
  HashMap UsedContexts;

  int SymbolCount;
  Bitstream BS;

  int CurrentContext;
  HuffmanTable CurrentHT;

  boolean WritingMode;
  boolean ReadingMode;

  HuffmanCoder (Bitstream BS)
  {
    this.BS = BS;
    SymStat = new SymbolStatistics ();
    Symbols = new GrowableIntArray ();
    Contexts = new GrowableIntArray ();
    UsedContexts = new HashMap();
    SymbolCount = 0;

    CurrentContext = 0;
    CurrentHT = null;

    WritingMode = false;
    ReadingMode = false;
  }

  void Encode (int Symbol, int Context)
  {
    // prohibit mixed encoding and decoding
    if( this.ReadingMode ) return;

    SymStat.AddSymbol (Symbol, Context);
    Symbols.Set (SymbolCount, Symbol);
    Contexts.Set (SymbolCount, Context);

    // just record which contexts are in use and set the associated huffman table to null.
    // it will be calculated and set later, when HuffmanCoder.Close() is called.
    UsedContexts.put( new Integer(Context), null );

    SymbolCount ++;
    WritingMode = true;
  }

  private void InitializeDecoding()
    throws java.io.IOException
  {
    // clear old contexts
    this.UsedContexts.clear();

    // read the huffman tables from the bit stream
    int context;
    int numContexts = this.BS.ReadInt();

    for (int i = 0; i < numContexts; ++i) {
      // read context number
      this.CurrentContext = this.BS.ReadInt();

      // decode a huffman table from the bitstream
      this.CurrentHT = new HuffmanTable();
      this.CurrentHT.Decode(this.BS);

      // update the huffman table association
      this.UsedContexts.put(new Integer(this.CurrentContext), this.CurrentHT);
    }

  }


  int Decode ( int context )
    throws java.io.IOException
  {
    // prohibit mixed encoding and decoding
    if( this.WritingMode ) return Integer.MIN_VALUE;

    if( !this.ReadingMode ) {
      this.InitializeDecoding();
      this.ReadingMode = true;
    }

    this.CurrentContext = context;
    this.CurrentHT = (HuffmanTable) this.UsedContexts.get( new Integer(this.CurrentContext) );

    // read bits from the stream until we find a matching symbol
    // in the current huffman table
    Code code = new Code();

      while (this.CurrentHT.FindCode(code) < 0)
        code.ReadBit(this.BS);

    int symbol = this.CurrentHT.FindCode( code );

    /*
    // check for a context switch
    if( symbol == this.CurrentHT.Codes.length-1 ) {
      // token for a context switch found

      // next integer value indicates next context to use
      this.CurrentContext = this.BS.ReadInt();
      this.CurrentHT = (HuffmanTable) this.UsedContexts.get( new Integer(this.CurrentContext) );

      return this.Decode();
    }
    */

    return symbol;
  }


  void Close ()
    throws java.io.IOException
  {
    if (WritingMode) {
      // if there are symbols to encode, write all the necessary information to the bit stream

      // encode the number of contexts (=number of encoded huffman tables)
      this.BS.WriteInt( this.UsedContexts.size() );

      // iterate through used contexts, build the huffman tables and write them to
      // the bit stream
      Iterator iterator = this.UsedContexts.keySet().iterator();

      while( iterator.hasNext() ) {
       this.CurrentContext = ((Integer) iterator.next()).intValue();

        // encode the context number
        this.BS.WriteInt(this.CurrentContext );

        // a new token (the largest value+1) is added to each context,
        // that will be used to indicate a context switch
        //this.SymStat.AddSymbol( SymStat.GetSymbolRange(this.CurrentContext ), this.CurrentContext );

        // calculate huffman table for this context
        this.CurrentHT = new HuffmanTable( this.SymStat,this.CurrentContext );
        this.CurrentHT.Encode( this.BS );

        // associate it with the appropriate number
        this.UsedContexts.put( new Integer(this.CurrentContext), this.CurrentHT );
      }

      // set up loop variables

      //int lastContext = this.Contexts.Get( 0 );
      //this.CurrentContext = this.Contexts.Get( 0 );
      //this.CurrentHT = (HuffmanTable) this.UsedContexts.get( new Integer(this.CurrentContext) );

      // first integer after the huffman table definitions indicates which huffman table
      // to start out with
      //this.BS.WriteInt( lastContext );

      for( int i = 0; i < this.SymbolCount; ++i ) {

       // see which context to use
       this.CurrentContext = this.Contexts.Get( i );
       this.CurrentHT = (HuffmanTable) this.UsedContexts.get( new Integer(this.CurrentContext) );

       /*
        // if the context changes, encode the context switch token (largest value in each huffman table)
        // followed by an integer value indicating which context to switch to
        if( lastContext !=this.CurrentContext ) {

          // encode context switch token (last one in each huffman table)
          // using the previous huffman table
          this.CurrentHT.GetCode( this.CurrentHT.Codes.length - 1 ).Encode( this.BS );

          // encode context to switch to
          this.BS.WriteInt(this.CurrentContext );

          // get the new huffman table
          this.CurrentHT = (HuffmanTable) this.UsedContexts.get( new Integer(this.CurrentContext) );
        }
        */
        // encode one symbol from the input stream
        this.CurrentHT.GetCode( this.Symbols.Get( i ) ).Encode( this.BS );

        // set current context to last context for next loop iteration
        //lastContext =this.CurrentContext;
      }
    }
    
    WritingMode = false;
    ReadingMode = false;
  }

  SymbolStatistics GetSymbolStatistics ()
  {
    return SymStat;
  }

  void WriteInt (int Number)
  {
    // dummy
    Symbols.Set (SymbolCount, Number);
    Contexts.Set (SymbolCount, 99);
    SymbolCount ++;
  }

  int ReadInt ()
    throws java.io.IOException
  {
    //dummy
    return BS.ReadInt ();
  }

  void Print ()
  {
      Iterator iterator = UsedContexts.keySet().iterator();

      while( iterator.hasNext() )
      {
        int context = ((Integer) iterator.next()).intValue();
        HuffmanTable HT = (HuffmanTable) this.UsedContexts.get (new Integer(context));
        System.out.println ("Kontext " + context + ":");
        HT.Print ();
      }
  }


  public static void main (String[] args)
    throws java.io.FileNotFoundException, java.io.IOException
  {
    Bitstream BSout = new Bitstream ("oumph.bs");

    HuffmanCoder huffmanCoder = new HuffmanCoder( BSout );
    int Symbols[] = {1, 2, 2, 3, 2, 4, 3, 4};
    for (int i = 0; i < 8; i ++)
        huffmanCoder.Encode (Symbols[i], i & 1); // Abwechselnd Kontext 0 und 1

    huffmanCoder.Close();

    System.out.println ("Huffman-Tables:");
    huffmanCoder.Print ();

    BSout.Close();

    Bitstream BSin = new Bitstream( "oumph.bs" );

    Quantized1D Result = new Quantized1D (8);
    HuffmanCoder huffmanCoder2 = new HuffmanCoder( BSin );
    for (int i = 0; i < 8; i ++)
        Result.Set (i, huffmanCoder2.Decode(i & 1));

    BSin.Close();

    System.out.println ("\nRekonstruierte Symbolfolge");
    Result.Print ();
  }
}
