import java.io.*;

class Signal1D extends Double1DArray
{
  private double Samples[];

  Signal1D (int Length)
  {
    Samples = new double[Length];
  }

  double Get (int n)
  {
    return Samples[n];
  }

  void Set (int n, double v)
  {
    Samples[n] = v;
  }

  int Length ()
  {
    return Samples.length;
  }

  final static int WAV_RIFF = (int)'R' + ((int)'I' << 8) + ((int)'F' << 16) + ((int)'F' << 24);
  final static int WAV_WAVE = (int)'W' + ((int)'A' << 8) + ((int)'V' << 16) + ((int)'E' << 24);
  final static int WAV_FMT  = (int)'f' + ((int)'m' << 8) + ((int)'t' << 16) + ((int)' ' << 24);
  final static int WAV_DATA = (int)'d' + ((int)'a' << 8) + ((int)'t' << 16) + ((int)'a' << 24);

  private static int Wav_Read (InputStream s, int bytes) throws IOException
  {
    int tmp = 0;
    for (int i = 0; i < bytes; i++)
        tmp += s.read() << (i*8);
    return tmp;
  }

  private static void Wav_Write (OutputStream s, int value, int bytes) throws IOException
  {
    for (int i = 0; i < bytes; i++)
    {   s.write (value & 255);
        value >>= 8;
    }
  }

  static Signal1D ReadWav (String FileName, int channel)
  {
        byte[] byteData = new byte[4];
        int[] sizeData = new int[4];
        Signal1D audio = null;
        
        try 
        {
            FileInputStream file = new FileInputStream (FileName);

            // RIFF header
            int riff = Wav_Read (file, 4);
            if (riff != WAV_RIFF)
               throw new IllegalArgumentException("no Wave-file");

            int size = Wav_Read (file, 4);
            int wave = Wav_Read (file, 4);
            if (wave != WAV_WAVE)
               throw new IllegalArgumentException("no Wave-file");

            // fmt chunk
            int fmtchunkid = Wav_Read (file, 4);
            if (fmtchunkid != WAV_FMT)
               throw new IllegalArgumentException("fmt chunk not found in wav file");

            int fmtchunksize   = Wav_Read (file, 4);
            int formattag      = Wav_Read (file, 2);
            int channelcount   = Wav_Read (file, 2);
            int samplerate     = Wav_Read (file, 4);
            int bytespersecond = Wav_Read (file, 4);
            int blockalign     = Wav_Read (file, 2);
            int bitspersample  = Wav_Read (file, 2);

            for (int i = 16; i < fmtchunksize; i ++)
                file.read ();

            // data chunk
            int datachunkid = Wav_Read (file, 4);
            if (datachunkid != WAV_DATA)
               throw new IllegalArgumentException("data chunk not found in wav file");
            int datachunklength = Wav_Read (file, 4);
            
            if (bitspersample == 16)
            {  audio = new Signal1D (datachunklength / 2 / channelcount);
               for (int i = 0; i < audio.Length (); i ++)
               {
                  for (int j = 0; j < channelcount; j ++)
                  {
                     int sample = Wav_Read (file, 2);
                     if (j == channel)
                        audio.Set (i, (double) sample);
                  }
               }
            }
            else
               throw new IllegalArgumentException("only 16 bits per sample allowed");

            file.close();
        }
        catch(IOException ex) 
        {
            System.out.println(ex);
        }
          
        return audio;
  }

  void WriteWav (String FileName)
  {
    WriteWav (FileName, 44100);
  }

  void WriteWav(String FileName, int samplerate)
  {
    try
    {
      FileOutputStream file = new FileOutputStream (FileName);

      // RIFF header
      Wav_Write (file, WAV_RIFF, 4);
      Wav_Write (file, Length() * 2 + 36, 4);
      Wav_Write (file, WAV_WAVE, 4);

      // fmt chunk
      Wav_Write (file, WAV_FMT, 4);
      Wav_Write (file, 16, 4);
      Wav_Write (file, 1, 2);
      Wav_Write (file, 1, 2); // channelcount
      Wav_Write (file, samplerate, 4);
      Wav_Write (file, samplerate*2, 4);
      Wav_Write (file, 2, 2);
      Wav_Write (file, 16, 2); // bitspersample

      // data chunk
      Wav_Write (file, WAV_DATA, 4);
      Wav_Write (file, Length() * 2, 4);
      
      for (int i = 0; i < Length (); i ++)
      {
          Wav_Write (file, (int) Math.round (Get (i)), 2);
      }

      file.close();
    }
    catch(IOException ex)
    {
      System.out.println(ex);
    }
  }

  void WaveletTransform (Filter Low, Filter High, int Depth)
  {
    Signal1D T = new Signal1D (Length ());

    int subArrayLength = Length();

    for(int i = 0; i < Depth; i++)
    {

      // Filtering is performed till the chosen depth is reached or the length of the
      // signal to be computed is shorter than
      // the length of the filter

      CopyTo(T);

      Low.Apply (new DoubleSubArray(T, 0, subArrayLength), new DoubleSubArray (this, 0, subArrayLength/2));
      High.Apply (new DoubleSubArray(T, 0, subArrayLength), new DoubleSubArray (this, subArrayLength/2, subArrayLength/2));

      subArrayLength /= 2;     
    }
  }
  

  void WaveletReconstruction (Filter Low, Filter High, int Depth)
  {
    int subArrayLength = Length() >> Depth;
   
    for(int i = 0; i < Depth; i++)
    {
      Signal1D T = new Signal1D (subArrayLength * 2);

      Low.InverseApply (new DoubleSubArray(this, 0, subArrayLength), T);
      High.InverseApply (new DoubleSubArray(this, subArrayLength, subArrayLength), T);

      subArrayLength *= 2;

      (new DoubleSubArray(this, 0, subArrayLength)).CopyFrom(new DoubleSubArray(T, 0, subArrayLength));

      T.SetAll(0);
    }  
  }

  // returned Signal is imaginary part
  Signal1D ComplexWaveletTransform (int Depth)
  {
    Signal1D I = new Signal1D (Length ());
    CopyTo (I);

    //dummy
    Filter L = Filter.NewLowHaar ();
    Filter H = Filter.NewHighHaar ();
    WaveletTransform (L, H, Depth);
    I.WaveletTransform (H, L, Depth);
    return I;
  }

  // returned Signal is imaginary part
  void ComplexWaveletReconstruction (Signal1D Imag, int Depth)
  {
    // dummy
    Filter L = Filter.NewLowHaar ();
    Filter H = Filter.NewHighHaar ();
    WaveletReconstruction (L, H, Depth);
    Imag.WaveletReconstruction (H, L, Depth);
    for (int i = 0; i < Length (); i ++)
        Set (i, (Get (i) + Imag.Get (i)) * 0.5);
  }

  public static void main (String[] args)
  {
     Signal1D S = new Signal1D (4);
     S.Set (0, 1.5);
     S.Set (1, 2.5);
     S.Set (2, 3.5);
     S.Set (3, 4.5);
     S.Print ();
     S.WaveletTransform (Filter.NewLowDaub4 (), Filter.NewHighDaub4 (), 2);
     S.Print ();
     S.WaveletReconstruction (Filter.NewLowDaub4 (), Filter.NewHighDaub4 (), 2);
     S.Print ();

     if (args.length < 2)
     {  System.out.println ("please specify a wav-filename");
        return;
     }

     Signal1D S2 = Signal1D.ReadWav (args[0], 1);
     // reverse it
     for (int i = 0; i < S2.Length () / 2; i ++)
     {   double tmp = S2.Get (i);
         S2.Set (i, S2.Get (S2.Length () - i - 1));
         S2.Set (S2.Length () - i - 1, tmp);
     }
     S2.WriteWav (args[1], 44100);
  }
}
